%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















:- ensure_loaded(typeops).

:- module_interface(decl).

:- import(prolog_types).
:- import(prolog_num).
:- import(prolog_io).
:- import(prolog_meta).

:- import(checking).
:- import(glib).
:- import_type(csppo,special_type).
:- import(cllib).    % tinf_diag/1

:- type symbol = top.

:- type typterm = top.   % type terms

:- type predkind --> (pred); (predl); (pred_neg); (rule).

%type preddecl --> pdecl(list(any),list(constraint)).
%type constraint --> any =< any.

:- pred pred_decl_variant_ngr0(symbol,nat,predkind,list(typterm),list(top)).
:- dynamic pred_decl_variant_ngr0/5.   
% pred_decl_variant_ngr0(PredName,Arity,PosNeg,ArgTypes,Constraints).
% PosNeg = pred if it is a standard, positive declaration
% PosNeg = pred_neg if it is a 'negated' declaration
% PosNeg = predl if it is a non-generic lemma predicate

:- pred func_decl_variant_ngr0(symbol,nat,list(typterm),typterm).
:- dynamic func_decl_variant_ngr0/4.   
% func_decl_variant_ngr0(FuncName,Arity,ArgTypes,TargetType).
% evtl. noch zusaetzlich FileId fuer reload/typecheck(File)


:- pred de_store_pdecl(symbol,nat,predkind,list(typterm),list(top)).
de_store_pdecl(PredName,Arity,PosNeg,ArgTypes,Constraints) :-
	assert(pred_decl_variant_ngr0(PredName,Arity,PosNeg,ArgTypes,Constraints)).

:- pred de_store_fdecl(symbol,nat,list(typterm),typterm).
de_store_fdecl(FuncName,Arity,ArgTypes,Target) :-
	assert(func_decl_variant_ngr0(FuncName,Arity,ArgTypes,Target)).

:- pred de_store_fdecl_goal(typterm,top,goal).
de_store_fdecl_goal(Type,Var,Goal) :-
        assert((func_decl_variant_ngr0(Var,_,[],Type) :- Goal) ).


:- pred func_decl_variant_ngr(symbol,list(typterm),typterm).
func_decl_variant_ngr(F,Args,Type) :-  true,
	func_decl_variant_ngr(F,_,Args,Type). % Arity unknown
	
:- pred func_decl_variant_ngr(symbol,nat,list(typterm),typterm).
func_decl_variant_ngr(F,Arity,Args,Type) :-  true,
        (func_decl_variant_ngr0(F,Arity,_,_) ->   % if at least one
	  func_decl_variant_ngr0(F,Arity,Args,Type);	    
	  (pragma_option(dummy_fdecl) ->
	   (gen_n_list(top,Arity,Args), % Args = [top,top, ...] as Arity
	    Type = top);
	   (write('!function declaration not found ! ':F/Arity),nl,
	    assert(tinf_diag(error_found)),
            false))
        ).

:- pred pred_decl_variant_ngr(top,list(top),list(top)).
pred_decl_variant_ngr(F,Args,Co) :-  true,
	pred_decl_variant_ngr(F,_,Args,Co).

:- pred pred_decl_variant_ngr(top,nat,list(top),list(top)).
pred_decl_variant_ngr(F,Arity,Args,Co) :-  
	true,
        ( pred_decl_variant_ngr0(F,Arity,_PosNeg,Args,Co) ->
	  true; 
	  (pragma_option(dummy_pdecl) ->
	   (gen_n_list(top,Arity,Args), % Args = [top,top, ...] as Arity
	    Co = []);
	   (write('!predicate declaration not found ! ':F/Arity),nl,
	    assert(tinf_diag(error_found)),
	    false))
        ).

:- pred pred_decl_lemma(top).
pred_decl_lemma(Head) :- % if declared with predl instead of pred
	% functor(Head,Arity,F), type-error
	functor(Head,F,Arity),
        pred_decl_variant_ngr0(F,Arity,KW,_Args,_Co),
	KW=predl.
	
:- pred test_pred_decl_variant_ngr(top,list(top),list(top)).
test_pred_decl_variant_ngr(F,Args,Co) :-  true,
        pred_decl_variant_ngr0(F,_,_,Args,Co).


:- pred test_pred_decl_variant_ngr(top,nat,list(top),list(top)).
test_pred_decl_variant_ngr(F,Arity,Args,Co) :-  true,
        pred_decl_variant_ngr0(F,Arity,_PosNeg,Args,Co).

/* currently unused
% declaration for pred or function symbol
pf_decl_variant_ngr(Symb,Args,Type,Co) :-
        ( pred_decl_variant_ngr0(Symb,_,Args,Co) ->
	    Type = goal
	;   Co=[], func_decl_variant_ngr(Symb,Args,Type)
	).
*/



declaration_variant_bottom(_,_,_) :- type_pragma(nocheck),
	true.
func_decl_variant(_,_,_) :- type_pragma(nocheck),
	true.
pred_decl_variant(_,_,_) :- type_pragma(nocheck),
	true.

:- pred pred_decl_distinct(symbol,list(_),list(_),list(nat)).
pred_decl_distinct(P,Args,Co,Ids) :-
	pred_decl_distinct(P,_,Args,Co,Ids).

:- pred pred_decl_distinct(symbol,nat,list(_),list(_),list(top)).
pred_decl_distinct(P,Arity,Args,Co,Ids) :-
        pred_decl_variant_ngr(P,Arity,Args,Co),
	reset_num_counter,
        my_number_tvars(Args,Ids1),
        my_number_tvars(Co,Ids2), % Also number existential vars in Co
        append(Ids1,Ids2,Ids).

% for simple genericity
:- pred pred_decl_generic(symbol,list(_),list(_),list(top)).
pred_decl_generic(P,Args,[],Ids) :-
	pred_decl_generic(P,_,Args,[],Ids).

% pred pred_decl_generic(symbol,nat,list(_),list(_),list(nat)). 
%			fuehrt su langen CSP-Zeiten bei typecheck von ti_cl.pl
:- pred pred_decl_generic(symbol,nat,list(_),list(_),list(top)).
pred_decl_generic(P,Arity,Args,[],Ids) :-
        pred_decl_variant_ngr(P,Arity,Args,Co),
	constraints_to_equiv(Co),  % ' T1 =< T2' implies T1 = T2
	reset_num_counter,
        my_number_tvars(Args,Ids).

:- pred constraints_to_equiv(list(top)).
constraints_to_equiv([]).
constraints_to_equiv([T1=<T2|Cs]) :-
	( var(T1) , var(T2) ->
	    true
	; write('Error: can not make generic':T1=<T2),
	  fail
	),
	T1 = T2,
	constraints_to_equiv(Cs).

/*
func_decl_variant_ngr_bot(Func,ArgList,TargetType) :-
	func_decl_variant_ngr_bot(Func,_,ArgList,TargetType).

func_decl_variant_ngr_bot(Func,Arity,ArgList,TargetType) :-
        func_decl_variant_ngr(Func,Arity,ArgList,TargetType),
        % replace variables of TargetType by bottom
        % if they do not occur in ArgList
        TargetType =.. [_Con|Param],
        place_bottom(Param,ArgList).
*/

:- pred  func_decl_variant_ngr_botvar(top,list(top),top,list(top)).
%
func_decl_variant_ngr_botvar(Func,ArgList,TargetType,BotVars) :-
	func_decl_variant_ngr_botvar(Func,_,ArgList,TargetType,BotVars).

:- pred  func_decl_variant_ngr_botvar(top,nat,list(top),top,list(top)).
%
func_decl_variant_ngr_botvar(Func,Arity,ArgList,TargetType,BotVars) :-
        func_decl_variant_ngr(Func,Arity,ArgList,TargetType),
        % collect variables of TargetType 
        % if they do not occur in ArgList
        TargetType =.. [_Con|Param],
        collect_botvar(Param,ArgList,BotVars).

/* unused
:- pred  pf_decl_variant_ngr_bot(symbol,list(top),top,list(top)).
pf_decl_variant_ngr_bot(Func,ArgList,TargetType,Co) :-
        pf_decl_variant_ngr(Func,ArgList,TargetType,Co),
        % replace variables of TargetType by bottom
        % if they do not occur in ArgList
        TargetType =.. [_Con|Param],
        place_bottom(Param,ArgList).
*/

:- pred place_bottom(list(top),list(top)).
place_bottom([],_).
place_bottom([X|Xs],Args) :- true,
        contains_var(Args,X), !,
        place_bottom(Xs,Args).
place_bottom([bottom|Xs],Args) :-
        place_bottom(Xs,Args).
 
:- pred collect_botvar(list(top),list(top),list(top)).
collect_botvar([],_,[]).
collect_botvar([X|Xs],Args,B) :- true,
        contains_var(Args,X), !,
        collect_botvar(Xs,Args,B).
collect_botvar([B|Xs],Args,[B|Bs]) :-
        collect_botvar(Xs,Args,Bs).

:- pred my_number_tvars(top,list(top)).
my_number_tvars(Any,[Any]) :- 
	var(Any),
	!,  
	new_num(N), 
	% Do not instantiate Any to a compund term,
	% otherwise it looks like a parametric type
	name(N,LN),         % inefficient but nice
	name(Any,[35|LN]).  % '#i' 
	% must correspond to pl2csp.pl !
my_number_tvars(Any,Ids) :- Any = [Arg|Args],!,
        my_number_tvars(Arg,Ids1),
        my_number_tvars(Args,Ids2),
        append(Ids1,Ids2,Ids).
my_number_tvars(Any,Ids) :-
        Any =.. [_F,Arg|Args],
        !,
        my_number_tvars([Arg|Args],Ids).
my_number_tvars(_,[]).



% transform predicate declaration: program syntax -> internal term
% @T -> LogVar
% T  -> New  with New=<T into constraint part
 
%decl_example(p(xx(T),lis(@T))).
% into  pdecl([xx(S),lis(T)],[S=<T]).

:- pred trans_pdecl_list(list(top),list(top),list(top)).
trans_pdecl_list([],[],[]).
trans_pdecl_list([X|Xs],[X2|X2s],Co) :-
	trans_pdecl(X,X2,Co1),
	trans_pdecl_list(Xs,X2s,Co2),
	append(Co1,Co2,Co).


:- pred trans_pdecl(top,top,list(top)).
trans_pdecl(V,TV,Co) :- var(V), !, TV = NewVar, Co = [ NewVar =< V ].
trans_pdecl(@V,TV,Co) :- !, true, TV = V, Co = [].
trans_pdecl(Term,TTerm,Co) :- % not var(Term) ; @V
	Term =.. [F|Args],
	trans_pdecl_list(Args,Args2,Co),
	TTerm =.. [F|Args2].

% trans_pdecl(Term,Term2) :- 
%	Term =.. [F|Args],
%	trans_pdecl_list(F,Args2),  % ill-typed !?
%	Term2 =.. [F|Args2].
%trans_pdecl(Term,Term2) :- % not var(Term)
%	Term =.. [F,Args],  % ill-typed !?
%	trans_pdecl_list(Args,Args2),
%	Term2 =.. [F|Args2].




% simplify_pdecl
%    constraint 'T =< T'   can be eliminated, where T is a variable
%    constraint 'S =< T'   can be eliminated, where S and T are variables
%                if S or T appear only once in (domain + constrints)
%    constraints 'S =< T' 'T =< U' can be replaced by 'S =< U', 
%                where T is a variable and occurs twice
%    constraint 'S =< T' can be eliminated if it is a exact '==' duplicate

% simplify_pdecl([A,B,int],[B=<C,int=<A,C=<xx],L).
%       L=[B=<xx,int=<A]

:- pred simplify_pdecl(list(top),list(top),list(top)).
simplify_pdecl(Domain,Co,Co_simp) :-
	simplify_pdecl_eq(Domain,Co,Co2),
	simplify_pdecl_single(Domain,Co2,Co3),
	simplify_pdecl_link(Domain,Co3,Co4),
	simplify_pdecl_dup(Co4,Co_simp).

:- pred simplify_pdecl_eq(list(top),list(top),list(top)).
simplify_pdecl_eq(_,[],[]).
simplify_pdecl_eq(_,[T1=<T2|Co],Co2) :-
	T1 == T2,
	!,
	simplify_pdecl_eq(_,Co,Co2).
simplify_pdecl_eq(_,[T1=<T2|Co],[T1=<T2|Co2]) :-
	simplify_pdecl_eq(_,Co,Co2).

:- pred simplify_pdecl_single(list(top),list(top),list(top)).
simplify_pdecl_single(_Domain,[],[]).
simplify_pdecl_single(Domain,Co,Co2) :-
	variables(Domain+Co,Vars,[]),
	simplify_pdecl_sv(Vars,Co,Co2).

:- pred simplify_pdecl_sv(list(top),list(top),list(top)).
simplify_pdecl_sv(_Vars,[],[]).
simplify_pdecl_sv(Vars,[T1=<T2|Co],Co2) :-
	( var_member_cnt(T1,Vars, 1)
	; var_member_cnt(T2,Vars, 1) ),
	!,
	simplify_pdecl_sv(Vars,Co,Co2).
simplify_pdecl_sv(Vars,[T1=<T2|Co],[T1=<T2|Co2]) :-
	simplify_pdecl_sv(Vars,Co,Co2).

:- pred simplify_pdecl_link(list(top),list(top),list(top)).
simplify_pdecl_link(_Domain,[],[]).
simplify_pdecl_link(Domain,Co,Co2) :-
	variables(Domain+Co,Vars,[]),
	simplify_pdecl_lv(Vars,Co,Co2).

:- pred simplify_pdecl_lv(list(top),list(top),list(top)).
simplify_pdecl_lv(_Vars,[],[]).
simplify_pdecl_lv(Vars,[T1=<T2|Co],Co2) :-
	special1_var_member_elim(T2=<T3,Co,Co1),
	var_member_cnt(T2,Vars, 2),
	!,
	simplify_pdecl_lv(Vars,[T1=<T3|Co1],Co2).
simplify_pdecl_lv(Vars,[T1=<T2|Co],Co2) :-
	special2_var_member_elim(T3=<T1,Co,Co1),
	var_member_cnt(T1,Vars, 2),
	!,
	simplify_pdecl_lv(Vars,[T3=<T2|Co1],Co2).
simplify_pdecl_lv(Vars,[T1=<T2|Co],[T1=<T2|Co2]) :-
	simplify_pdecl_lv(Vars,Co,Co2).

:- pred simplify_pdecl_dup(top,list(top)).
simplify_pdecl_dup(CD,C) :- var_remove_duplicates(CD,C).


:- pred special1_var_member_elim(top,list(top),list(top)).
special1_var_member_elim(X=<Y,[X1=<Y|T],R) :- X1==X, !, R=T.
special1_var_member_elim(X,[Y|T],[Y|R]) :- special1_var_member_elim(X,T,R).

:- pred special2_var_member_elim(top,list(top),list(top)).
special2_var_member_elim(X=<Y,[X=<Y1|T],R) :- Y1==Y, !, R=T.
special2_var_member_elim(X,[Y|T],[Y|R]) :- special2_var_member_elim(X,T,R).


:- pred var_member_cnt(top,list(top),nat).
var_member_cnt(Var,List,N) :-
	var_member_cnt_a(Var,List,0,N).

:- pred var_member_cnt_a(top,list(top),nat,nat).
var_member_cnt_a(_, [], Acc, Acc).
var_member_cnt_a(Var, [Head|L], Acc, N) :-
        Head == Var,
        !,
	Acc1 is Acc + 1,
	var_member_cnt_a(Var,L,Acc1,N).
var_member_cnt_a(Var, [_|Tail], Acc, N) :-
        var_member_cnt_a(Var, Tail, Acc, N).


% SICStus Prolog Manual p.176
:- pred variables(top,list(top),list(top)).
variables(X, [X|L0], L) :- var(X), !, L = L0.
variables(T, L0, L) :-
	functor(T, _, A),
	variables(0, A, T, L0, L).

:- pred variables(nat,nat,top,list(top),list(top)).
variables(A, A, _, L0, L) :- !, L = L0.
variables(A0, A, T, L0, L) :-
%	A0 < A,
	A1 is A0+1,
	arg(A1, T, X),
	variables(X, L0, L1),
	variables(A1, A, T, L1, L).


%-----
% gen_n_list(Elem,N,List) :-
%       generate a list of length N,
%       each element of List equals Elem
:- pred gen_n_list(@T,nat,list(@T)).
%
gen_n_list(_,0,[]).
gen_n_list(Elem,N,[Elem|L]) :-
        N > 0,
        N1 is N -1,
        gen_n_list(Elem,N1,L).

:- end_interface(decl).

:- module_body(decl).

:- pred pragma_option(top).

:- end_module(decl).
