%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















:- ensure_loaded(typeops). 

:- module_interface(ti_clause).

%
%  ti_clause takes a Horn clause or query clause and generates a 
%  corresponding type-inference problem, i.e. a  structure of type 'clc'
%

:- import(prolog_types).

:- import(glib).
:- import(csppo).
:- import(ti_cl).

:- op(600,xfy,(#)).  % s.a. ti_cl.pl 


%-----------------------------------------------------
% ti_clause(Lat, clause, CLC).
%       result CLC: clause constraints
%		clc(Lat,
%		    Red,     % normalized reduction, value-var : type-term
%		    PUMs     % constraints for sim.p.u.m.
%		    HC       % constraints of head predicate,
%		    Co       % type constraints
%		    B        % Variables which may be bottom
%
% Atom # g   well-typed Atom + genericity/head condition
% Atom # o   well-typed Atom
% Term * T   Term has least-type T
% Term : T   Term has type T
%

:- type task --> 
	clause(top,list(top)) ;   % Head and list of subgoals
	query(list(top)).         % list of subgoals

:- type status --> nocheck ; more.

:- type tpragma --> type_pragma(top).

% type inferencing 

:- pred ti_clause(poset(_), task, status, clc).

ti_clause(Lat, clause(H,Goals) , Status, CLC) :- !,
	true,
	(   member(type_pragma(nocheck),Goals) -> Status=nocheck
	; 
	    make_f_pairs_const(#,Goals,o,G2),
	    (   member_elim(type_pragma(non_generic)#o,G2,G4) ->
	        ti_cl(Lat, [H#o|G4], CLC)
	    ;   pred_decl_lemma(H) ->
	        ti_cl(Lat, [H#o|G2], CLC)
	    ;   ti_cl(Lat, [H#g|G2], CLC)  % genericity/head condition
	    ),
	    Status=more
	).

ti_clause(Lat, query(Goals) ,Status, CLC) :- !,  % Goal
	true,
	(   member(type_pragma(nocheck),Goals) -> Status=nocheck
	;
	    make_f_pairs_const(#,Goals,o,G2),
	    ti_cl(Lat, G2, CLC),
	    Status=more
	).

:- end_interface(ti_clause).

:- module_body(ti_clause).

:- end_module(ti_clause).
