%----
% Copyright (C) 1996, Gregor Meyer, FernUniversitaet Hagen, Germany.
% This file is part of the system 'Typical'.
% See the file LICENSE for conditions of distribution and use.
%----
%:- multifile (subtype)/1, (suptype)/1, (type)/1, (pred)/1, (type_import)/1, (local)/1.
















%
% :- module tinf.
%
% type inference for Prolog clauses.
%

:- ensure_loaded(typeops). 

:- module_interface(tinf).

:- import(prolog_types).
:- import(prolog_num).
:- import(prolog_meta).

%:- import(poset).
:- import(tap).
:- import(glib).
:- import(csppo).
:- import(ti_2po).
:- import(ti_clause).  % task
:- import(ti_cl).
:- import(tinf_more).

:- type prolog_clause = top.

%-----------------------------------------------------
% type inference

:- pred type_infer_cl(poset(_),list(prolog_clause)). 

type_infer_cl(_,[]).
type_infer_cl(Lat,[Cl|Cls]) :-
        numbervars(Cl,0,_),
	retractall(tinf_diag(_)),   % see cllib.pl
/*      get0(XXX),
        (XXX=120 -> write(escaped), get0(_CR), fail
        ; true
        ),
        \+ XXX = 120, % ='x'
*/
        (pragma_option(verbose) ->
	 (nl,write('check/infer: '),write(Cl),nl);
	    true),
        (true
        ;true,fail),
        %
        type_inference_nv(Lat,Cl,_Co,VT), !,
        (pragma_option(verbose) ->
	 (write('Consistent with variable types '),write(VT),nl);
	    true),
	%  write(co:_Co),nl,
        type_infer_cl(Lat,Cls).
type_infer_cl(Lat,[_Ill|Cls]) :-
	% error message already written
        write('Press Enter to continue'),
        get0(_XXX),
        type_infer_cl(Lat,Cls).
 

:- pred type_inference(poset(_),prolog_clause,list(top),list(top)). 
type_inference(Lat, Clause, C,VT) :-
	numbervars(Clause,0,_),  % translates X to '$VAR'(I)
	type_inference_nv(Lat, Clause, C,VT).

:- pred type_inference_nv(poset(_),prolog_clause,list(top),list(top)). 
type_inference_nv(Lat, Clause, C,VT) :-
	Lat1=Lat,
	reset_num_counter,    % for numbering type parameters
	(Clause = ( H :- Body) ->
	    commas_to_list(Body,BL),
	    Task = clause(H,BL)
	;   Clause = ( :- Body) ->
	    commas_to_list(Body,BL),
	    Task = query(BL)
	;   Task = clause(Clause,[])   % Clause is single fact
	),
	type_inf0(Lat1,Task,C,VT).

/* old variant
type_inference_nv(Lat, Clause, C,VT) :-
	(option_with_bottom -> 
	 complete_bottom(Lat,Lat1); 
	    Lat1=Lat),
	reset_num_counter,    % for numbering type parameters
	(Clause = ( H :- Body) ->
	    commas_to_list(Body,BL),
	    Task = clause(H,BL)
	;   Clause = ( :- Body) ->
	    commas_to_list(Body,BL),
	    Task = query(BL)
	;   Task = clause(Clause,[])   % Clause is single fact
	),
	type_inf0(Lat1,Task,C,VT).

*/
:- pred type_inf0(poset(_),task,list(top),list(top)).
type_inf0(Lat,Task,C,VT) :-
	type_inf(Lat,Task,C,VT),
	!.
type_inf0(Lat,Task,_C,_VT) :-
	% normal type-inf failed, now make error analysis
	\+ tinf_diag(error_found),   % may already be found in decl.pl
	type_inference_ea(Lat,Task,_C,_VT),
	fail.

% error analysis
%  errors are detected straight forward
% on error first test head, then
% try to put step by step the subgoals into the body and 
% test each 
% 
:- pred type_inference_ea(poset(_),task,list(top),list(top)).
%
%type_inference_ea(_Lat,CL,_C,_VT) :-
%   write(ti_ea(CL)),nl,fail.

type_inference_ea(Lat,clause(H,[]),C,VT) :-
	(   type_inf(Lat,clause(H,[]),C,VT) ->
	    true
	;
	    nl, write('type error caused by head':H),nl,
	    fail
	).
    
% process normal clauses with body: 
%
type_inference_ea(Lat,clause(H,Body),C,VT) :-
	append(B,[LastGoal],Body),
	type_inference_ea(Lat,clause(H,B),_C0,VT0),
	(   type_inf(Lat,clause(H,Body),C,VT) ->
	    true
	;
	    write('The clause '), write_prolog(clause(H,B)),nl,
	    write('is type-consistent using VT='),write(VT0),nl,
	    write('With additional subgoal '),write(LastGoal),nl,
	    write('the clause '),write_prolog(clause(H,Body)),nl, 
	    write('is not type-consistent.'),nl,
	    type_inference_ea_args(Lat,clause(H,B),LastGoal),
	    fail
	).


% process queries

type_inference_ea(_Lat,query([]),[],[]).

type_inference_ea(Lat,query(Body),C,VT) :-
	append(B,[LastGoal],Body),
	type_inference_ea(Lat,query(B),_C0,VT0),
        (   (Body=[];type_inf(Lat,query(Body),C,VT)) ->  
            true
        ; 
	    (   B=[_|_] ->
	        write('The clause: '),write_prolog(query(B)), nl,
	        write('is type-consistent using VT='),write(VT0),nl,
	        write('With additional subgoal '),write(LastGoal),nl
	    ;   true
	    ),
	    write('the clause: '),write_prolog(query(Body)), nl,
	    write('is not type-consistent.'),nl,
	    type_inference_ea_args(Lat,query(B),LastGoal),
	    fail
	).


 
% Analyse arguments of erroneous subgoal
%     generalize a subset of the arguments in LastGoal
%     then try if clause is type-consistent with generalized subgoal
%
:- pred type_inference_ea_args(poset(_),task,top).
%
type_inference_ea_args(Lat,Clause0,LastGoal) :-
	% generalize a subset of the arguments in LastGoal
        anti_subst(LastGoal,_Subst,MoreGeneral), 
        % anti_subst/3 may generate MoreGeneral as a variable
        nonvar(MoreGeneral), 
        % write(anti_subst(LastGoal,_Subst,MoreGeneral)), nl,
        numbervars(MoreGeneral,104,_),  % Var104 written as 'A4'
        % write('trying '),write(MoreGeneral),nl,
	( Clause0 = clause(H,B) ->
            append(B,[MoreGeneral],Body),
	    Clause = clause(H,Body)
	; Clause0 = query(B),
            append(B,[MoreGeneral],Body),
	    Clause = query(Body)
	),
        % write(type_inf(_,Clause,_C,VT)),nl,
        type_inf(Lat,Clause,_C,VT),    
        write('The clause with the last atom GENERALIZED to  '), 
        write(MoreGeneral) ,nl,
        write('would be type-consistent using VT='),write(VT),nl,nl,
        !.
type_inference_ea_args(_,_,_).  % catch all
 

:- pred type_inf(poset(_),task,list(top),list(top)).
type_inf(Lat,Task,C,VT) :-
        ti_clause(Lat,Task, Status, CLC),
        ( true;
          true,fail),
        true, % nocheck; more; error?
        (    Status = nocheck -> write(nocheck),nl
        ;    true,
             type_inf_more(CLC,C,VT)
        ).
	
:- end_interface(tinf).

:- module_body(tinf).

:- pred pragma_option(top).

:- pred commas_to_list(prolog_clause,list(top)).
 
commas_to_list((A,B),[A|BL]) :- !, commas_to_list(B,BL).
commas_to_list(A,[A]).
 

% write_prolog(clause(H,B)) :-
% write_prolog(query(B)) :-
%	write argument as a Prolog clause, i.e., elements of B separated by ','
:- pred write_prolog(prolog_clause).
%
write_prolog(clause(H,[])) :-  
	write(H),
	write('.').
write_prolog(clause(H,[B|Bs])) :- 
	write(H), 
	write(' :- '), 
	write(B),
	write_comma_list(Bs),
	write('.').
write_prolog(query([B|Bs])) :-
	write(' :- '), 
	write(B),
	write_comma_list(Bs),
	write('.').
	

:- pred write_comma_list(list(top)).
%
write_comma_list([]).
write_comma_list([B|Bs]) :-
	write(', '),
	write(B),
	write_comma_list(Bs).

:- end_module(tinf).


